package org.xbib.interlibrary.ldap;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.naming.NamingEnumeration;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import javax.naming.directory.SearchResult;
import org.xbib.datastructures.json.tiny.Json;
import org.xbib.datastructures.tiny.TinyMap;
import org.xbib.interlibrary.api.library.LibraryGroups;
import org.xbib.interlibrary.ldap.util.LdapClient;
import org.xbib.interlibrary.ldap.util.SearchScope;
import org.xbib.settings.Settings;

public class LdapLibraryGroups implements LibraryGroups {

    private static final Logger logger = Logger.getLogger(LdapLibraryGroups.class.getName());

    private final Settings settings;

    private final LdapClient ldapClient;

    public LdapLibraryGroups(Settings settings) {
        this.settings = settings;
        this.ldapClient = new LdapClient(settings);
    }

    @Override
    public Collection<String> getAllGroups() {
        return Arrays.asList(settings.getAsArray("allgroups"));
    }

    @SuppressWarnings("unchecked")
    @Override
    public Collection<String> getMembers(String group) {
        final List<String> list = new ArrayList<>();
        ldapClient.search(String.format(settings.get("ldap.getmembers.basedn"), group),
        SearchScope.valueOf(settings.get("ldap.getmembers.scope").toUpperCase()),
        String.format(settings.get("ldap.getmembers.filter"), group),
        null,
        results -> {
            try {
                while (results != null && results.hasMore()) {
                    SearchResult sr = results.next();
                    String dn = sr.getNameInNamespace();
                    Attributes attrs = sr.getAttributes();
                    NamingEnumeration<? extends Attribute> namingEnumeration = attrs.getAll();
                    TinyMap.Builder<String, Object> map = TinyMap.builder();
                    map.put("dn", dn);
                    while (namingEnumeration.hasMore()) {
                        Attribute attr = namingEnumeration.next();
                        String key = attr.getID();
                        if (attr.size() == 1) {
                            map.put(key, attr.get());
                        } else {
                            List<Object> l = new ArrayList<>();
                            for (int i = 0; i < attr.size(); ++i) {
                                l.add(attr.get(i));
                            }
                            map.put(key, l);
                        }
                    }
                    for (Map.Entry<String, Object> entry : map.build().entrySet()) {
                        if ("memberUid".equals(entry.getKey())) {
                            Object o = entry.getValue();
                            if (o instanceof List) {
                                list.addAll((List<String>) o);
                            } else {
                                list.add((String) o);
                            }
                        }
                    }
                }
            } catch (Exception e) {
                logger.log(Level.WARNING, e.getMessage(), e);
            }
        });
        return list;
    }

    @SuppressWarnings("unchecked")
    @Override
    public Collection<String> getGroups(String isil) throws IOException {
        List<String> groups = new ArrayList<>();
        ldapClient.search(settings.get("ldap.getgroup.basedn"),
            SearchScope.valueOf(settings.get("ldap.getgroup.scope").toUpperCase()),
            String.format(settings.get("ldap.getgroup.filter"), isil),
            null,
            results -> {
                try {
                    while (results != null && results.hasMore()) {
                        SearchResult sr = results.next();
                        String dn = sr.getNameInNamespace();
                        Attributes attrs = sr.getAttributes();
                        NamingEnumeration<? extends Attribute> namingEnumeration = attrs.getAll();
                        TinyMap.Builder<String, Object> map = TinyMap.builder();
                        map.put("dn", dn);
                        while (namingEnumeration.hasMore()) {
                            Attribute attr = namingEnumeration.next();
                            String key = attr.getID();
                            if (attr.size() == 1) {
                                map.put(key, attr.get());
                            } else {
                                List<Object> l = new ArrayList<>();
                                for (int i = 0; i < attr.size(); ++i) {
                                    l.add(attr.get(i));
                                }
                                map.put(key, l);
                            }
                        }
                        for (Map.Entry<String, Object> entry : map.build().entrySet()) {
                            if ("cn".equals(entry.getKey())) {
                                groups.add((String) entry.getValue());
                            } else if ("description".equals(entry.getKey())) {
                                Map<String, Object> m = Json.toMap((String) entry.getValue());
                                if (m.containsKey("role")) {
                                    Object o = m.get("role");
                                    if (o instanceof Iterable) {
                                        for (String s : (Iterable<String>) o) {
                                            groups.add(s);
                                        }
                                    } else {
                                        groups.add((String) o);
                                    }
                                }
                            } else {
                                if ("role".equals(entry.getKey())) {
                                    Object o = entry.getValue();
                                    if (o instanceof Iterable) {
                                        for (String s : (Iterable<String>) o) {
                                            groups.add(s);
                                        }
                                    } else {
                                        groups.add((String) o);
                                    }
                                }
                            }
                        }
                    }
                } catch (Exception e) {
                    logger.log(Level.WARNING, e.getMessage(), e);
                }
            });
        return groups;
    }

    @Override
    public boolean addGroup(String isil, String group) throws IOException {
        if (getMembers(group).contains(isil)) {
            return false;
        }
        Map<String, Object> add = Map.of("memberUid", isil);
        Map<String, Object> replace = Map.of();
        Map<String, Object> remove = Map.of();
        String dn = String.format(settings.get("ldap.getmembers.basedn"), group);
        ldapClient.modify(dn, add, replace, remove);
        logger.log(Level.INFO, "added group " + group + " to ISIL " + isil);
        return true;
    }

    @Override
    public boolean removeGroup(String isil, String group) throws IOException {
        if (!getMembers(group).contains(isil)) {
            return false;
        }
        Map<String, Object> add = Map.of();
        Map<String, Object> replace = Map.of();
        Map<String, Object> remove = Map.of("memberUid", isil);
        String dn = String.format(settings.get("ldap.getmembers.basedn"), group);
        ldapClient.modify(dn, add, replace, remove);
        logger.log(Level.INFO, "removed group " + group + " from ISIL " + isil);
        return true;
    }
}
