package org.xbib.interlibrary.ldap;

import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Objects;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.naming.NamingEnumeration;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import javax.naming.directory.SearchResult;
import javax.naming.ldap.Rdn;
import org.xbib.datastructures.tiny.TinyMap;
import org.xbib.interlibrary.api.library.DefaultStation;
import org.xbib.interlibrary.api.library.LibraryLocations;
import org.xbib.interlibrary.api.library.Station;
import org.xbib.interlibrary.ldap.util.LdapClient;
import org.xbib.interlibrary.ldap.util.SearchScope;
import org.xbib.settings.Settings;

public class LdapLibraryLocations implements LibraryLocations {

    private static final Logger logger = Logger.getLogger(LdapLibraryLocations.class.getName());

    private final Settings settings;

    private final LdapClient ldapClient;

    public LdapLibraryLocations(Settings settings) {
        this.settings = settings;
        this.ldapClient = new LdapClient(settings);
    }

    @Override
    public List<String> getAccounts() throws IOException {
        return getAccounts(settings.get("ldap.basedn"), settings.get("ldap.filter"));
    }

    @Override
    public List<String> getConfiguredAccounts() throws IOException {
        String basedn = settings.get("ldap.people.basedn");
        Objects.requireNonNull(basedn, "ldap.people.basedn must not be null");
        String filter = settings.get("ldap.people.filter");
        Objects.requireNonNull(basedn, "ldap.people.filter must not be null");
        return getAccounts(basedn, filter).stream().filter(s -> s.startsWith("DE-")).toList();
    }

    public List<String> getAccounts(String basedn, String filter) throws IOException {
        Objects.requireNonNull(basedn, "basedn must not be null");
        Objects.requireNonNull(filter, "filter must not be null");
        List<String> list = new ArrayList<>();
        ldapClient.search(
                basedn,
                SearchScope.SUB,
                filter,
                "description",
                results -> {
                    try {
                        while (results != null && results.hasMore()) {
                            SearchResult sr = results.next();
                            Attributes attrs = sr.getAttributes();
                            NamingEnumeration<? extends Attribute> namingEnumeration = attrs.getAll();
                            while (namingEnumeration.hasMore()) {
                                Attribute attr = namingEnumeration.next();
                                String key = attr.getID();
                                if ("uid".equals(key)) {
                                    if (attr.size() == 1) {
                                        list.add(attr.get().toString());
                                    } else {
                                        for (int i = 0; i < attr.size(); ++i) {
                                            list.add(attr.get(i).toString());
                                        }
                                    }
                                }
                            }
                        }
                    } catch (Exception e) {
                        logger.log(Level.WARNING, e.getMessage(), e);
                    }
                });
        return list;
    }

    @SuppressWarnings("unchecked")
    @Override
    public List<String> getLocations(String isil, EnumSet<Station.Process> processes) throws IOException {
        List<String> list = new ArrayList<>();
        for (Station.Process process : processes) {
            ldapClient.search(String.format(settings.get("ldap.location.basedn"), isil, process.name().toLowerCase(Locale.ROOT)),
                    SearchScope.SUB,
                    "objectclass=*",
                    "description",
                    results -> {
                        try {
                            while (results != null && results.hasMore()) {
                                SearchResult sr = results.next();
                                String dn = sr.getNameInNamespace();
                                Attributes attrs = sr.getAttributes();
                                NamingEnumeration<? extends Attribute> namingEnumeration = attrs.getAll();
                                TinyMap.Builder<String, Object> map = TinyMap.builder();
                                map.put("dn", dn);
                                while (namingEnumeration.hasMore()) {
                                    Attribute attr = namingEnumeration.next();
                                    String key = attr.getID();
                                    if (attr.size() == 1) {
                                        map.put(key, attr.get());
                                    } else {
                                        List<Object> l = new ArrayList<>();
                                        for (int i = 0; i < attr.size(); ++i) {
                                            l.add(attr.get(i));
                                        }
                                        map.put(key, l);
                                    }
                                }
                                for (Map.Entry<String, Object> entry : map.build().entrySet()) {
                                    if ("l".equals(entry.getKey())) {
                                        Object o = entry.getValue();
                                        if (o instanceof List) {
                                            list.addAll((List<String>) o);
                                        } else {
                                            list.add((String) o);
                                        }
                                    }
                                }
                            }
                        } catch (Exception e) {
                            logger.log(Level.WARNING, e.getMessage(), e);
                        }
                    });
        }
        return list;
    }

    @Override
    public void addLocation(String isil, String location, Station.Process process) {
        // first add organization if not present
        String organizationDN = settings.get("ldap.location.basedn");
        ldapClient.add(String.format(organizationDN, isil, process.name().toLowerCase(Locale.ROOT)),
                Map.of("objectClass", "organization", "l", location));
        String locationDN = "l=%s," + organizationDN;
        ldapClient.add(String.format(locationDN, escapeDN(location), isil, process.name().toLowerCase(Locale.ROOT)),
                Map.of("objectClass", "locality", "l", location));
    }

    @Override
    public void removeLocation(String isil, String location, Station.Process process) throws IOException {
        String organizationDN = settings.get("ldap.location.basedn");
        String locationDN = "l=%s," + organizationDN;
        ldapClient.remove(String.format(locationDN, escapeDN(location), isil, process.name().toLowerCase(Locale.ROOT)));
    }

    @SuppressWarnings("unchecked")
    @Override
    public List<Station> getStations(String isil) throws IOException {
        List<Station> list = new ArrayList<>();
        EnumSet.allOf(Station.Process.class).forEach(p -> {
            ldapClient.search(String.format(settings.get("ldap.location.basedn"), isil, p.name().toLowerCase(Locale.ROOT)),
                    SearchScope.SUB,
                    "objectclass=*",
                    "description",
                    results -> {
                        try {
                            String stationIsil = null;
                            List<URI> uris = new ArrayList<>();
                            List<String> localities = new ArrayList<>();
                            List<String> sortkeys = new ArrayList<>();
                            while (results != null && results.hasMore()) {
                                SearchResult sr = results.next();
                                String dn = sr.getNameInNamespace();
                                Attributes attrs = sr.getAttributes();
                                NamingEnumeration<? extends Attribute> namingEnumeration = attrs.getAll();
                                // step 1: build a tiny map
                                TinyMap.Builder<String, Object> builder = TinyMap.builder();
                                builder.put("dn", dn);
                                while (namingEnumeration.hasMore()) {
                                    Attribute attr = namingEnumeration.next();
                                    String key = attr.getID();
                                    if (attr.size() == 1) {
                                        builder.put(key, attr.get());
                                    } else {
                                        List<Object> l = new ArrayList<>();
                                        for (int i = 0; i < attr.size(); ++i) {
                                            l.add(attr.get(i));
                                        }
                                        builder.put(key, l);
                                    }
                                }
                                Map<String, Object> map = builder.build();
                                // step 2, collect attributes from o DN's
                                if (dn.startsWith("o")) {
                                    // in desktop we have o, not uid
                                    if (map.containsKey("o")) {
                                        stationIsil = map.get("o").toString();
                                    }
                                    // if not desktop, we may have user ids
                                    if (map.containsKey("uid")) {
                                        stationIsil = map.get("uid").toString();
                                    }
                                    // we misuse "st" (state) as our container for custom URIs
                                    if (builder.containsKey("st")) {
                                        Object o = builder.get("st");
                                        if (o instanceof List) {
                                            uris.addAll(((List<String>) o).stream().map(URI::create).toList());
                                        } else {
                                            uris.addAll(Arrays.stream(o.toString().split(",")).map(URI::create).toList());
                                        }
                                    }
                                }
                                if (dn.startsWith("l")) {
                                    if (map.containsKey("l")) {
                                        Object o = map.get("l");
                                        if (o instanceof List) {
                                            localities.addAll((List<String>) o);
                                        } else {
                                            localities.add((String) o);
                                        }
                                    }
                                    if (map.containsKey("description")) {
                                        sortkeys.add(map.get("description").toString());
                                    }
                                }
                            }
                            // step 3 after all results have been processed, iterate over localities, and build stations.
                            // - first locality is the preset locality
                            // - only BCS2 and UPLOAD are INPUT type, all other are OUTPUT
                            // - default circulation is COPY
                            for (int i = 0; i < localities.size(); i++) {
                                String locality = localities.get(i);
                                String sortkey = sortkeys.get(i);
                                Station station = DefaultStation.builder()
                                        .setEnabled(true)
                                        .setProcess(p)
                                        .setISIL(stationIsil)
                                        .setName(locality)
                                        .setSortKey(sortkey)
                                        .setURIs(addStationIsilTo(uris, stationIsil))
                                        .build();
                                list.add(station);
                            }
                        } catch (Exception e) {
                            logger.log(Level.WARNING, e.getMessage(), e);
                        }
                    });
        });
        return list;
    }

    @Override
    public void addStation(Station station) {
        if (station == null) {
            return;
        }
        String process = station.getProcess().name().toLowerCase(Locale.ROOT);
        // first URI determines host
        URI uri = station.getURIs().isEmpty() ? URI.create("sftp://localhost") : station.getURIs().getFirst();
        createLibraryEntry(ldapClient, process, station.getISIL(), uri.getHost());
        String dn = String.format(settings.get("ldap.location.basedn"), station.getISIL(), process);
        String locality = station.getName();
        String description = station.getSortKey();
        createLocalityEntry(ldapClient, "l=" + escapeDN(locality) + dn, locality, description);
    }

    @Override
    public void removeStation(Station station) {
        String process = station.getProcess().name().toLowerCase(Locale.ROOT);
        String baseDN = String.format(settings.get("ldap.location.basedn"), station.getISIL(), process);
        String locality = station.getName();
        ldapClient.remove("l=" + escapeDN(locality) + baseDN);
        // do not remove the baseDN
    }

    private List<URI> addStationIsilTo(List<URI> list, String isil) {
        if (list == null) {
            return null;
        }
        return list.stream().map(u -> URI.create(u.toString().replace("://", "://" + isil + "@"))).toList();
    }

    private void createLibraryEntry(LdapClient ldapClient, String process, String isil, String host) {
        String dn = String.format(settings.get("ldap.location.basedn"), isil, process);
        logger.log(Level.INFO, "dn=" + dn + " isil=" + isil);
        Map<String, Object> m = new HashMap<>();
        switch (process) {
            case "bcs2" -> {
                List<String> list = Arrays.asList(settings.getAsArray("bcs2." + isil));
                if (!list.isEmpty()) {
                    m.put("objectClass", List.of("organization", "account", "posixAccount"));
                    m.put("cn", isil);
                    m.put("uid", "S-" + isil);
                    m.put("uidNumber", list.get(0));
                    m.put("gidNumber", list.get(1));
                    m.put("gecos", "sftp://" + host + ":22/" + list.get(2) + "/IMG/");
                    m.put("homeDirectory", "/data/fl/" + isil);
                    m.put("userPassword", findPassword(ldapClient, "uid=S-" + isil + ",ou=People,o=Fernleihe,dc=hbz-nrw,dc=de"));
                }
            }
            case "upload" -> {
                List<String> list = Arrays.asList(settings.getAsArray("upload." + isil));
                if (!list.isEmpty()) {
                    m.put("objectClass", List.of("organization", "account", "posixAccount"));
                    m.put("cn", isil);
                    m.put("uid", "S-" + isil);
                    m.put("uidNumber", list.get(0));
                    m.put("gidNumber", list.get(1));
                    m.put("gecos", "sftp://" + host + ":22/" + list.get(2) + "/UPLOAD/");
                    m.put("homeDirectory", "/data/fl/" + isil);
                    m.put("userPassword", findPassword(ldapClient, "uid=S-" + isil + ",ou=People,o=Fernleihe,dc=hbz-nrw,dc=de"));
                }
            }
            case "m3print" -> {
                List<String> list = Arrays.asList(settings.getAsArray("m3print." + isil));
                if (!list.isEmpty()) {
                    m.put("objectClass", List.of("organization", "account", "posixAccount"));
                    m.put("cn", isil);
                    m.put("uid", "P-" + isil);
                    m.put("uidNumber", list.get(0));
                    m.put("gidNumber", list.get(1));
                    m.put("gecos", "sftp://" + host + ":22/" + list.get(2) + "/SPOOL/DELV/," +
                            "sftp://" + host + ":22/" + list.get(2) + "/SPOOL/FRONTSH/," +
                            "sftp://" + host + ":22/" + list.get(2) + "/SPOOL/RL/");
                    m.put("homeDirectory", "/data/fl/" + isil);
                    m.put("userPassword", findPassword(ldapClient, "uid=P-" + isil + ",ou=People,o=Fernleihe,dc=hbz-nrw,dc=de"));
                }
            }
            default -> {
                m.put("objectClass", "organization");
            }
        }
        m.put("o", isil);
        ldapClient.add(dn, m);
    }

    private String findPassword(LdapClient ldapClient, String baseDN) {
        StringBuilder password = new StringBuilder();
        ldapClient.search(baseDN,
                SearchScope.ONE,
                "(&(objectclass=posixAccount))",
                null,
                results -> {
                    try {
                        TinyMap.Builder<String, Object> builder = TinyMap.builder();
                        while (results != null && results.hasMore()) {
                            SearchResult sr = results.next();
                            NamingEnumeration<? extends Attribute> namingEnumeration = sr.getAttributes().getAll();
                            while (namingEnumeration.hasMore()) {
                                Attribute attr = namingEnumeration.next();
                                String key = attr.getID();
                                if (attr.size() == 1) {
                                    builder.put(key, attr.get());
                                } else {
                                    List<Object> l = new ArrayList<>();
                                    for (int i = 0; i < attr.size(); ++i) {
                                        l.add(attr.get(i));
                                    }
                                    builder.put(key, l);
                                }
                            }
                        }
                        Map<String, Object> map = builder.build();
                        password.append(map.containsKey("userPassword") ? map.get("userPassword").toString() : null);
                    } catch (Exception e) {
                        logger.log(Level.WARNING, e.getMessage(), e);
                    }
                });
        return password.toString();
    }

    private void createLocalityEntry(LdapClient ldapClient, String dn, String locality, String description) {
        logger.log(Level.INFO, "dn=" + dn + " locality=" + locality + " description=" + description);
        ldapClient.add(dn, Map.of("objectClass", "locality", "l", locality, "description", description));
    }

    private static String escapeDN(String string) {
        if (string == null) {
            return null;
        }
        return Rdn.escapeValue(string).replaceAll("/", "\\\\/");
    }
}
