package org.xbib.interlibrary.ldap.util;

import java.io.Closeable;
import java.io.IOException;
import java.util.Map;
import java.util.Objects;
import java.util.Properties;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Consumer;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.naming.CommunicationException;
import javax.naming.Context;
import javax.naming.NameNotFoundException;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.DirContext;
import javax.naming.directory.SearchResult;
import javax.naming.ldap.Control;
import javax.naming.ldap.InitialLdapContext;
import javax.naming.ldap.LdapContext;
import javax.naming.ldap.SortControl;
import javax.naming.ldap.SortResponseControl;

import org.xbib.settings.Settings;

public class LdapClient implements Closeable {

    private static final Logger logger = Logger.getLogger(LdapClient.class.getName());

    private final Settings settings;

    private final AtomicBoolean closed;

    private LdapContext ctx;

    public LdapClient(Settings settings) {
        this.settings = settings;
        this.closed = new AtomicBoolean(true);
    }

    @Override
    public void close() throws IOException {
        if (closed.compareAndExchange(false, true)) {
            close(ctx);
        }
    }

    public synchronized void search(String base,
                                    SearchScope scope,
                                    String filter,
                                    String sortKey,
                                    Consumer<NamingEnumeration<SearchResult>> consumer) {
        ensureOpenContext();
        NamingEnumeration<SearchResult> results = null;
        try {
            Search search = new Search(base, scope, filter);
            if (sortKey != null) {
                ctx.setRequestControls(new Control[]{new SortControl(sortKey, Control.NONCRITICAL)});
            }
            Control[] controls = ctx.getResponseControls();
            if (controls != null) {
                for (Control control : controls) {
                    if (control instanceof SortResponseControl sortResponseControl) {
                        if (!sortResponseControl.isSorted()) {
                            throw sortResponseControl.getException();
                        }
                    }
                }
            }
            logger.log(Level.FINE, "LDAP Client: search base = " + base + " filter = " + filter + " scope = " + scope);
            results = ctx.search(search.getBase(), search.getFilter(), null, search.createSearchControls());
            consumer.accept(results);
        } catch (CommunicationException e) {
            logger.log(Level.WARNING, e.getMessage());
            close(ctx);
            ensureOpenContext();
        } catch (NameNotFoundException e) {
            // ignore
        } catch (NamingException | IOException e) {
            throw new IllegalStateException(e);
        } finally {
            try {
                if (results != null) {
                    results.close();
                }
            } catch (NamingException e) {
                logger.log(Level.FINEST, e.getMessage());
            }
        }
    }

    public synchronized void add(String dn,
                                 Map<String, Object> entry) {
        ensureOpenContext();
        DirContext dirContext = null;
        try {
            Add add = new Add(dn, entry);
            logger.log(Level.FINE, "LDAP Client: add name = " + dn + " attributes = " + entry);
            dirContext = ctx.createSubcontext(add.getName(), add.getAttributes());
        } catch (CommunicationException e) {
            logger.log(Level.WARNING, e.getMessage());
            close(ctx);
            ensureOpenContext();
        } catch (NamingException e) {
            logger.log(Level.WARNING, e.getMessage(), e);
        } finally {
            close(dirContext);
        }
    }

    public void modify(String dn,
                       Map<String, Object> add,
                       Map<String, Object> replace,
                       Map<String, Object> remove) {
        ensureOpenContext();
        try {
            Modify modify = new Modify(dn, add, replace, remove);
            logger.log(Level.FINE, "LDAP Client: modify name = " + dn + " add = " + add + " modify = " + modify + " remove = " + remove);
            ctx.modifyAttributes(modify.getName(), modify.getModificationItems());
        } catch (CommunicationException e) {
            logger.log(Level.WARNING, e.getMessage());
            close(ctx);
            ensureOpenContext();
        } catch (NamingException e) {
            logger.log(Level.WARNING, e.getMessage(), e);
        }
    }

    public void remove(String dn) {
        try {
            Remove remove = new Remove(dn);
            ctx.destroySubcontext(remove.getName());
        } catch (CommunicationException e) {
            logger.log(Level.WARNING, e.getMessage());
            close(ctx);
            ensureOpenContext();
        } catch (NamingException e) {
            logger.log(Level.WARNING, e.getMessage(), e);
        }
    }

    private void ensureOpenContext() {
        if (closed.compareAndSet(true, false)) {
            this.ctx = createContext(settings);
        }
    }

    private static void close(Context context) {
        if (context == null) {
            return;
        }
        try {
            context.close();
        } catch (NamingException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
    }

    private static LdapContext createContext(Settings settings) {
        try {
            Properties properties = new Properties();
            properties.setProperty(Context.INITIAL_CONTEXT_FACTORY, "com.sun.jndi.ldap.LdapCtxFactory");
            properties.setProperty(Context.PROVIDER_URL,
                    Objects.requireNonNull(settings.get("ldap.url"), "ldap.url must not be null"));
            properties.setProperty(Context.SECURITY_AUTHENTICATION, "simple");
            properties.setProperty(Context.SECURITY_PRINCIPAL,
                    Objects.requireNonNull(settings.get("ldap.user"), "ldap.user must not be null"));
            properties.setProperty(Context.SECURITY_CREDENTIALS,
                    Objects.requireNonNull(settings.get("ldap.password"), "ldap.password must not be null"));
            properties.setProperty(Context.REFERRAL, "ignore");
            properties.setProperty("com.sun.jndi.ldap.connect.timeout",
                    settings.get("ldap.timeout.connect", "10000"));
            properties.setProperty("com.sun.jndi.ldap.read.timeout",
                    settings.get("ldap.timeout.read","10000"));
            properties.setProperty("com.sun.jndi.ldap.connect.pool",
                    settings.get("ldap.pool.enabled", "true"));
            return new InitialLdapContext(properties, null);
        } catch (NamingException e) {
            throw new IllegalStateException(e);
        }
    }
}
